@import url('theme.css');

/* Logo background */
.wy-side-nav-search, .wy-side-nav-search img {
    background-color: #ffffff !important;
}

.highlight {
  background: #f1f3f4;
}

.navbar {
  background: #ffffff;
}

.navbar-nav {
  background: #ffffff;
}

/* side bar */
.wy-nav-side {
  background: #f1f3f4;
}

.wy-menu-vertical a {
  color: #707070;
}

.wy-side-nav-search div.version {
  color: #404040;
}
