#!/usr/bin/env bash

## Update the following line to test changes to CI images
## See https://xgboost.readthedocs.io/en/latest/contrib/ci.html#making-changes-to-ci-containers

# IMAGE_TAG=main
# or PR number:
# IMAGE_TAG=PR-88

IMAGE_TAG=main
