## Get details for AWS ECR (Elastic Container Registry) in environment variables

ECR_AWS_ACCOUNT_ID="492475357299"
ECR_AWS_REGION="us-west-2"
DOCKER_REGISTRY_URL="${ECR_AWS_ACCOUNT_ID}.dkr.ecr.${ECR_AWS_REGION}.amazonaws.com"
